import datetime
import os
import pandas as pd

import crawler

#=====================================================================
# Define constants
#=====================================================================

ENDPOINT_URL = "https://api.trafikinfo.trafikverket.se/v2/data.json"
DEFAULT_AUTHKEY = '8067eee1cc894ceda6a1c635b4543731'
DEBUG = False
OUTPUT_VERSION = 2 #output format version

COLUMNS_TRAINANNOUNCEMENT = [   
    "ActivityId",
    "ActivityType",
    "Advertised",
    "AdvertisedTimeAtLocation",
    "AdvertisedTrainIdent",
    "Booking",
    "Canceled",
    "Deleted",
    "Deviation",
    "EstimatedTimeAtLocation",
    "EstimatedTimeIsPreliminary",
    "FromLocation",
    "InformationOwner",
    "LocationSignature",
    "MobileWebLink",
    "ModifiedTime",
    "NewEquipment",
    "Operator",
    "OtherInformation",
    "PlannedEstimatedTimeAtLocation",
    "PlannedEstimatedTimeAtLocationIsValid",
    "ProductInformation",
    "ScheduledDepartureDateTime",
    "Service",
    "TechnicalDateTime",
    "TechnicalTrainIdent",
    "TimeAtLocation",
    "TimeAtLocationWithSeconds",
    "ToLocation",
    "TrackAtLocation",
    "TrainComposition",
    "TrainOwner",
    "TypeOfTraffic",
    "WebLink",
    "WebLinkName",
    "ViaFromLocation",
    "ViaToLocation"
]

COLUMNS_TRAINMESSAGE = [
    "CountyNo",
    "Deleted",
    "EndDateTime",
    "EventId",
    "ExternalDescription",
    "Geometry",
    "Header",
    "LastUpdateDateTime",
    "ModifiedTime",
    "PrognosticatedEndDateTimeTrafficImpact",
    "ReasonCode",
    "StartDateTime",
    "TrafficImpact"
]

#=====================================================================
# Trafikverket API Query building
#=====================================================================

def format_datetime(dt, utc=False):
    """Format a datetime object as in Trafikverkets API

    Args:
        dt (datetime.datetime): The datetime object to format
        utc (bool, optional): Use UTC format instead of local time. Defaults to False (local time).

    Returns:
        str: The formatted string
    """

    if utc:
        return dt.astimezone(tz=datetime.timezone.utc).isoformat(timespec='seconds').replace("+00:00","Z")
    else:
        return dt.astimezone().isoformat(timespec='seconds')

def build_query(authkey, attributes = {}, filter = "", columns = []):
    attributestr = " ".join(["{key}='{value}'".format(key=key, value=value) for key, value in attributes.items()])    
    includestr = "\n".join(["<INCLUDE>{column}</INCLUDE>".format(column=column) for column in columns])

    query = """
    <REQUEST>
        <LOGIN authenticationkey='{authkey}' />
        <QUERY {attributestr}>
            <FILTER>
                {filter}
            </FILTER>
            {includestr}
        </QUERY>
    </REQUEST>
    """.format(authkey=authkey, attributestr = attributestr, filter=filter, includestr=includestr)

    return query

def get_train_announcement_query(authkey, startTime = None, endTime=None, locationSignature= "", sseurl=False, limit=None):
    attributes = {
        "objecttype": "TrainAnnouncement",
        "schemaversion": "1.6",
        "orderby": "TimeAtLocation",
        "includedeletedobjects": "true",
        "sseurl": sseurl
    }
    if not limit is None:
        attributes["limit"] = limit

    filter = """
        <EQ name="LocationSignature" value="{locationSignature}"/>
    """.format(locationSignature=locationSignature)

    if not startTime is None:
        filter += """
            <OR>
                <GTE name="AdvertisedTimeAtLocation" value="{startTime}" />
                <GTE name="TimeAtLocation" value="{startTime}" />
            </OR>
        """.format(startTime=format_datetime(startTime))
    
    if not endTime is None:
        filter += """
            <OR>
                <LT name="AdvertisedTimeAtLocation" value="{endTime}" />
                <LT name="TimeAtLocation" value="{endTime}" />
            </OR>
        """.format(endTime=format_datetime(endTime))
    
    return build_query(authkey, attributes=attributes,filter=filter, columns= COLUMNS_TRAINANNOUNCEMENT)

def get_train_message_query(authkey, startTime = None, endTime=None, locationSignature= "", sseurl=False, limit=None):
    attributes = {
        "objecttype": "TrainMessage",
        "schemaversion": "1.7",
        "orderby": "StartDateTime",
        "includedeletedobjects": "true",
        "sseurl": sseurl
    }
    if not limit is None:
        attributes["limit"] = limit

    filter = """
        <IN name="TrafficImpact.AffectedLocation.LocationSignature" value="{locationSignature}"/>
    """.format(locationSignature=locationSignature)

    if not startTime is None:
        filter += """<GTE name="StartDateTime" value="{startTime}" />\n""".format(startTime=format_datetime(startTime))

    if not endTime is None:
        filter += """<LT name="StartDateTime" value="{endTime}" />\n""".format(endTime=format_datetime(endTime))

    return build_query(authkey, attributes=attributes,filter=filter, columns= COLUMNS_TRAINMESSAGE)


#=====================================================================
# Daily Trafikverket crawlers
#=====================================================================

class DailyTrainAnnouncementCrawler (crawler.PeriodicCrawler):
    """Crawler that daily fetches all train announcements (arrival, departure) for the previous day for a given location.
    """
    locationSignature = ""

    def __init__(self, path, authkey, interval=60*60, locationSignature="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from api.trafikinfo.trafikverket.se
            interval (float, optional): Interval in seconds that the crawler checks if data needs to be crawled. Defaults to 60*60.
            locationSignature (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(ENDPOINT_URL, path, authkey, interval=interval, version=version, output_version=OUTPUT_VERSION)
        self.locationSignature = locationSignature

    def crawl(self):
        todayMidnight = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        yesterdayMidnight = todayMidnight - datetime.timedelta(days=1)
        yesterdaysDataCompleteTime = datetime.datetime.now().replace(hour=3, minute=0, second=0, microsecond=0) #yesterdays ata assumed complete at 3:00


        filename = os.path.join(self.path, "TrainAnnouncement_" + self.locationSignature + "_" + yesterdayMidnight.strftime("%Y-%m-%d") + ".csv")

        if not os.path.exists(filename) and datetime.datetime.now() > yesterdaysDataCompleteTime:
            self.log_status("Crawling yesterdays data...")

            #Fetch data for yesterday
            query = get_train_announcement_query(self.authkey, startTime=yesterdayMidnight, endTime=todayMidnight, locationSignature=self.locationSignature)

            try:
                data = self._query(query)

                #check for error message
                try:
                    self.log_warning(data["RESPONSE"]["RESULT"][0]["ERROR"])
                except Exception:
                    pass

                #get data
                try:
                    data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0]["TrainAnnouncement"], columns = COLUMNS_TRAINANNOUNCEMENT)
                    data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])

                    if self.DEBUG:
                        self.log_info("Received {rows} rows.".format(rows=len(data)))
                except Exception as e:
                    self.log_exception(e)

                #write to csv
                data.to_csv(filename, index=False)
                self.writes += 1
                if self.DEBUG:
                    self.log_info("Written {rows} rows.".format(rows=len(data)))
            except Exception as e:
                self.log_exception(e)
        else:
            if self.DEBUG:
                self.log_info("No need to crawl at this point.")



class DailyTrainMessageCrawler (crawler.PeriodicCrawler):
    """Crawler that daily fetches all train messages (disruption messages) for the previous day for a given location.
    """
    locationSignature = ""

    def __init__(self, path, authkey, interval=60*60, locationSignature="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from api.trafikinfo.trafikverket.se
            interval (float, optional): Interval in seconds that the crawler checks if data needs to be crawled. Defaults to 60*60.
            locationSignature (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(ENDPOINT_URL, path, authkey, interval=interval, version=version, output_version=OUTPUT_VERSION)
        self.locationSignature = locationSignature

    def crawl(self):
        todayMidnight = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        yesterdayMidnight = todayMidnight - datetime.timedelta(days=1)

        filename = os.path.join(self.path, "TrainMessage_" + self.locationSignature + "_" + yesterdayMidnight.strftime("%Y-%m-%d") + ".csv")
        
        if not os.path.exists(filename):
            if self.DEBUG:
                self.log_status("Crawling yesterdays data...")

            #Fetch data for yesterday
            query = get_train_message_query(self.authkey, startTime=yesterdayMidnight, endTime=todayMidnight, locationSignature=self.locationSignature)
            
            try:
                data = self._query(query)

                #check for error message
                try:
                    self.log_warning(data["RESPONSE"]["RESULT"][0]["ERROR"])
                except Exception:
                    pass

                #get data
                try:
                    data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0]["TrainMessage"], columns = COLUMNS_TRAINMESSAGE)
                    data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])

                    if self.DEBUG:
                        self.log_info("Received {rows} rows.".format(rows=len(data)))
                except Exception as e:
                    self.log_exception(e)

                #write to csv
                data.to_csv(filename, index=False)
                self.writes += 1
                if self.DEBUG:
                    self.log_info("Written {rows} rows.".format(rows=len(data)))
            except Exception as e:
                self.log_exception(e)
        else:
            if self.DEBUG:
                self.log_info("Latest daily data already written. No need to crawl at this point.")

#=====================================================================
# Streaming Trafikverket crawlers
#=====================================================================

class StreamTrainAnnouncementCrawler (crawler.StreamCrawler):
    """Crawler that continously monitors all train announcements (arrival, departure) for a given location.
    """
    def __init__(self, path, authkey, locationSignature="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from api.trafikinfo.trafikverket.se
            locationSignature (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(crawler.trafikverket.ENDPOINT_URL, path, authkey, version=version, output_version=OUTPUT_VERSION)
        self.locationSignature = locationSignature
        self.inactivity_seconds = 3*60*60 #restart connection if no events received for 3 hours

    @property
    def fileprefix(self):
        return "TrainAnnouncement_" + self.locationSignature + "_"

    def _get_query(self):
        return get_train_announcement_query(self.authkey, locationSignature=self.locationSignature, sseurl=True, limit=0)
    
    def _handle_events(self, data):
        #check for error message
        try:
            ex = Exception(data["RESPONSE"]["RESULT"][0]["ERROR"])
            self.log_exception(ex)
        except Exception:
            pass

        #get data
        try:
            data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0]["TrainAnnouncement"], columns = COLUMNS_TRAINANNOUNCEMENT)
            data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])
            if self.DEBUG:
                self.log_info("Received {rows} rows.".format(rows=len(data)))
            #write to csv
            data.to_csv(self.csvfile, index=False, mode='a', header=(not os.path.exists(self.csvfile)))
            if self.DEBUG:
                self.log_info("Written {rows} rows.".format(rows=len(data)))
            self.writes += 1
        except Exception as e:
            self.log_exception(e)

class StreamTrainMessageCrawler (crawler.StreamCrawler):
    """Crawler that continously monitors all train messages (disruption messages) for a given location.
    """
    def __init__(self, path, authkey, locationSignature="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from api.trafikinfo.trafikverket.se
            locationSignature (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(crawler.trafikverket.ENDPOINT_URL, path, authkey, version=version, output_version=OUTPUT_VERSION)
        self.locationSignature = locationSignature
        self.inactivity_seconds = 10*60*60 #restart connection if no events received for 10 hours

    @property
    def fileprefix(self):
        return "TrainMessage_" + self.locationSignature + "_"

    def _get_query(self):
        return get_train_message_query(self.authkey, locationSignature=self.locationSignature, sseurl=True, limit=0)
    
    def _handle_events(self, data):
        #check for error message
        try:
            ex = Exception(data["RESPONSE"]["RESULT"][0]["ERROR"])
            self.log_exception(ex)
        except Exception:
            pass

        #get data
        try:
            data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0]["TrainMessage"], columns = COLUMNS_TRAINMESSAGE)
            data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])
            
            if self.DEBUG:
                self.log_info("Received {rows} rows.".format(rows=len(data)))
            #write to csv
            data.to_csv(self.csvfile, index=False, mode='a', header=(not os.path.exists(self.csvfile)))
            if self.DEBUG:
                self.log_info("Written {rows} rows.".format(rows=len(data)))
            self.writes += 1
        except Exception as e:
            self.log_exception(e)

